# coding: utf-8
# (倉庫番 sōkoban, warehouse keeper)
# Created in 1981 by Hiroyuki Imabayashi
# Published in 1982 by Thinking Rabbit
# Based on the Ruby Quiz #5 implementation
# http://blade.nagaokaut.ac.jp/cgi-bin/scat.rb/ruby/ruby-talk/118577
# Copyright 2004 Dennis Ranke <dennis.ranke at epost.de>
# Copyright 2013 Boohbah <boohbah at gmail.com>

require File.expand_path('../../lib/sokoban/version.rb', __FILE__)
require 'io/console'
require 'optparse'
require 'logger'
require 'zlib'

module Sokoban
  HOME = ENV['HOME']
  BASENAME = File.basename($0, '.rb')
  SAV_FILE = "#{HOME}/sokoban.save.gz"
  LOG_FILE = "#{HOME}/sokoban-#{Time.now.to_i}.log"
  LVL_FILE = File.expand_path('../../sokoban_levels.txt', __FILE__)

  LEVELS = File.open(LVL_FILE).readlines.map do |l|
    l.chomp.ljust(19)
  end.join("\n").split(/\n {19}\n/).map do |l|
    l.gsub(/\n/, '')
  end

  class Game
    def initialize
      @level = Level.new
      parse_opts
      # --level and --resume opts may change lnum
      @lnum = @level.lnum
      play
    end

  private

    def play
      loop do
        begin
          @lnum += 1 if @level.play
        rescue RuntimeError
          # Write log in case of quit
          log_moves
          exit
        end
        log_moves
        break if @lnum > 50
        @level.restart(@lnum)
      end
      puts "You won Sokoban!"
      puts "やったね！"
    end

    def log_moves
      @logger.debug @level.inspect if @options[:debug]
    end

    def parse_opts
      @options = {}
      @optparse = OptionParser.new do |opts|
        opts.banner = "Usage: #{BASENAME} [options]\n"
        opts.banner << "Controls:
     wasd, hjkl, number pad          Move
     ^C                              Quit without save
     b                               Toggle batch mode
     r                               Restart
     u                               Undo
     q                               Quit\n"
        opts.banner << "Options:"

        opts.on("-h", "--help", "Display help") do
          puts opts
          exit
        end

        opts.on("-v", "--version", "Print version number") do
          puts "#{BASENAME} #{VERSION}"
          exit
        end

        opts.on("-b", "--batch", "Batch input mode") do
          $batch = true
        end

        opts.on("-d", "--debug", "Enable debug log") do
          @logger = Logger.new LOG_FILE
          @logger.level = Logger::DEBUG
          @logger.debug "Created logger"
          @logger.debug "#{BASENAME} #{VERSION}"
          @logger.debug "#{$0} --debug #{ARGV.join(" ")}"
          @options[:debug] = true
        end

        opts.on("-l N", "--level N", Integer, "Start at level N") do |n|
          @level = Level.new(n)
          @options[:level] = true
        end
        
        opts.on("-r", "--resume", "Resume saved game") do
          @level.load_game
          @options[:resume] = true
        end

        opts.on("-x") do
          $hax = true
        end
      end

      begin
        @optparse.parse!
        if @options[:level] && @options[:resume]
          raise ArgumentError, "Can't set --level and --resume. Choose one."
        end
      rescue OptionParser::InvalidOption,
             OptionParser::InvalidArgument,
             OptionParser::MissingArgument,
             RangeError, ArgumentError
        puts "#{@optparse}\n"
        puts "ERROR: #{$!}"
        exit
      end
    end
  end

  class Level
    attr_reader :lnum, :pushes

    def initialize(lnum = 1)
      unless lnum.between?(1, 50)
        raise RangeError, "Level out of range. Choose 1-50"
      end
      @lnum = lnum
      @moves = []
      @level = LEVELS[@lnum - 1].ljust(19 * 16)
      @state = [@level.dup]
      @pushes = 0
    end

    def play
      while free_boxes > 0
        print self
        if $batch
          # Input a string terminated by newline
          parse_move(gets.strip)
        else
          # Handle single key press
          parse_move(STDIN.getch)
        end
        if $hax
          break if @moves.size > 3 # Hax
        end
      end
      print self
      print "Level #{@lnum} complete! "
      unless @lnum == 50
        print "Congratulations, on to level #{@lnum + 1}. "
      end
      STDIN.getch
    end

    def parse_move(input)
      # Handle multiple moves
      input.each_char {|c| parse_move(c) } if input.size > 1
      case input.downcase
      when 'b'; $batch = ! $batch
      when 'w', 'k', '8'; move(:W) # up
      when 'a', 'h', '4'; move(:A) # left
      when 's', 'j', '2'; move(:S) # down
      when 'd', 'l', '6'; move(:D) # right
      when 'r'; initialize(@lnum)
      when 'u'; undo
      when 'q'; quit
      when "\u0003"; puts; exit # ^C - Quit without save 
      end
    end

    def free_boxes
      @level.scan(/o/).size
    end

    def moves
      @moves.size
    end

    # The public interface to initialize()
    def restart(lnum)
      initialize(lnum)
    end

    def inspect
      s = "Level: #{@lnum} Moves: #{moves} Pushes: #{@pushes} "
      s << @moves.join
    end

    def to_s
      clear_screen
      s = "Level: #{@lnum}\n\n"
      s << (0...16).map {|i| @level[i * 19, 19] }.join("\n")
      s << "\nMoves: #{moves} Pushes: #{@pushes}\n> "
    end

    def load_game
      file = Zlib::GzipReader.open(SAV_FILE) {|f| f.read }
      initialize(file.slice!(/\d+/).to_i)  # Init @lnum
      file.each_char {|c| move(c.to_sym) } # Rebuild @moves
    end

  private

    def save_game
      file = ([@lnum] << @moves).flatten.join
      Zlib::GzipWriter.open(SAV_FILE) {|f| f.write(file) }
    end

    def clear_screen
      30.times { puts }
    end

    def undo
      if @moves.size > 0
        @moves.pop
        @state.pop
        @pushes -= 1 if box_moved?
        @level = @state.last.dup
      end
    end

    def box_moved?
      @level.split(//).each_with_index do |char, i|
        if char =~ /o|\*/
          return true if @state.last[i] != char
        end
      end
      return false
    end

    def quit
      print "Save game? [y/n] "
      puts c = STDIN.getch
      case c
      when 'y', 'Y'
        print "Saving... "
        save_game
      end
      puts "Goodbye"
      # rescue RuntimeError in Game class
      raise RuntimeError
    end

    def find_player
      pos = @level.index(/@|\+/)
      return pos % 19, pos / 19
    end

    def [](x, y)
      @level[x + y * 19]
    end

    def []=(x, y, v)
      @level[x + y * 19] = v
    end

    DIRECTIONS = {
      W: { x:  0, y: -1 },
      A: { x: -1, y:  0 },
      S: { x:  0, y:  1 },
      D: { x:  1, y:  0 }
    }

    def move(direction)
      x, y = find_player
      dx = DIRECTIONS[direction][:x]
      dy = DIRECTIONS[direction][:y]
      dest = self[x + dx, y + dy]

      case dest
      when '#' # This is a wall, no move
        return
      when 'o', '*'
        # Check dest2 behind box, same direction times 2
        dest2 = self[x + dx * 2, y + dy * 2]
        if dest2 == ' '
          self[x + dx * 2, y + dy * 2] = 'o'
        elsif dest2 == '.'
          self[x + dx * 2, y + dy * 2] = '*'
        else # Destination is blocked, no move
          return
        end
        # Update previous box position
        dest = (dest == 'o') ? ' ' : '.'
        @pushes += 1
      end

      # Update player position
      self[x + dx, y + dy] = (dest == ' ') ? '@' : '+'
      # Update previous position
      self[x, y] = (self[x, y] == '@') ? ' ' : '.'
      @moves << direction.to_s
      # Save a copy of the level in case of undo
      @state << @level.dup
    end
  end
end
